'use strict';

import BaseError from './types/base-error';
import ErrorFormatter from './error-formatter';

/**
 *
 */
export default class ErrorResponseFormatter {

  /**
   * Create a new ErrorResponseHandler instance.
   */
  constructor() {
    this._formatters = [];
  }

  /**
   * Check if a formatter has been applied.
   *
   * @param {Error} errorType
   * @returns {boolean}
   */
  hasFormatterForType(errorType) {
    let hasFormatter = false;
    this._formatters.forEach(function(f) {
      if (f.formatType() === errorType) {
        hasFormatter = true;
        return;
      }
    });
    return hasFormatter;
  }

  /**
   * Remove a formatter for a specific type of error.
   *
   * @param {Error} errorType
   * @returns {boolean}
   */
  removeForType(errorType) {
    let formatter = null;
    this._formatters.forEach(function(f) {
      if (f.formatType() === errorType) {
        formatter = f;
        return;
      }
    });

    if (formatter) {
      this._formatters.splice(this._formatters.indexOf(formatter), 1);
    }

    return (formatter !== null);
  }

  /**
   * Apply an {ErrorFormatter} to the response handler.
   *
   * @param {ErrorFormatter} errorFormatter
   */
  apply(errorFormatter) {
    if (errorFormatter instanceof ErrorFormatter) {
      if (this._formatters.indexOf(errorFormatter) > -1) {
        throw new Error('The specified error formatter has already been applied');
      } else {
        if (this.hasFormatterForType(errorFormatter.formatType())) {
          throw new Error('A formatter of the same error type has already been applied');
        } else {
          this._formatters.push(errorFormatter);
        }
      }
    } else {
      throw new Error('{errorFormatter} is not a valid instance of {ErrorFormatter}');
    }
  }

  /**
   * Format a particular type of error.
   *
   * @param {Error} error
   * @returns {object}
   */
  format(error) {
    if (error instanceof BaseError) {
      delete error.status;
    }

    this._formatters.forEach(function(f) {
      if (error instanceof f.formatType()) {
        error = f.format(error);
      }
    });

    return error;
  }

}
