'use strict';

import fs from 'fs';
import path from 'path';
import cors from 'cors';
import IO from 'socket.io';
import helmet from 'helmet';
import logger from 'morgan';
import Express from 'express';
import Context from './context';
import ErrorTypes from './errors';
import Config from 'quture-config';
import Utilities from './utilities';
import favicon from 'serve-favicon';
import bodyParser from 'body-parser';
import RandomString from 'randomstring';
import cookieParser from 'cookie-parser';
import ErrorCode from './errors/error-code';
import methodOverride from 'method-override';
import {default as createDebugger} from 'debug';
import ErrorFormatter from './errors/error-formatter';
import ErrorResponseFormatter from './errors/error-response-formatter';

const debug = createDebugger('quture:app');

/**
 *
 */
export default class Quture {

  /**
   * Create a new {Quture} instance.
   *
   * @param {object} options
   */
  constructor(options) {
    options = options || {};

    if (!options.rootDir) {
      throw new Error('{options.rootDir} is required');
    } else {
      if (fs.lstatSync(options.rootDir).isDirectory()) {
        this._rootDir = options.rootDir;
      } else {
        throw new Error('{options.rootDir} does not exist');
      }
    }

    /**
     * Configure the configuration handler
     */
    if (!options.config) {
      if (options.configDir && typeof options.configDir !== 'string') {
        throw new Error('{options.configDir} must be of type string');
      }

      let configDir = options.configDir || path.join(this._rootDir, 'config');

      Utilities.ensureDirSync(configDir);

      options.config = new Config({ configDir });
    } else {
      if (!(options.config instanceof Config)) {
        throw new Error('{options.config} must be a valid instance of {Config}');
      }
    }

    this._config = options.config;

    let appConfigFile = path.join(__dirname, 'config', 'app.js');

    if (!this.isConfigurationFilePublished(this._config, appConfigFile)) {
      this.publishConfigurationFile(this._config, appConfigFile);
    }

    /**
     * Setup views directory
     */
    if (options.viewsDir) {
      this._viewsDir = options.viewsDir;
    } else {
      this._viewsDir = path.join(this.getRootDirectory(), 'views');
    }

    Utilities.ensureDirSync(this._viewsDir);

    /**
     * Setup public directory
     */
    if (options.publicDir) {
      this._publicDir = options.publicDir;
    } else {
      this._publicDir = path.join(this.getRootDirectory(), 'public');
    }

    Utilities.ensureDirSync(this._publicDir);

    /**
     * Setup storage directory
     */
    if (options.storageDir) {
      this._storageDir = options.storageDir;
    } else {
      this._storageDir = path.join(this.getRootDirectory(), 'storage');
    }

    Utilities.ensureDirSync(this._storageDir);

    /**
     * Setup Websockets
     */
    this._io = IO();

    /**
     * Setup Error Response Formatter
     */
    this._errorResponseFormatter = new ErrorResponseFormatter();

    /**
     * Contextualize the instance as a global context.
     */
    Quture.contextualize(this, 'global');

    /**
     * Create a new router
     */
    this._router = Quture.createRouter();

    /**
     * //
     */
    this._isBootstrapped = false;
  }

  /**
   * Retrieve the utilities helper class.
   *
   * @return {Utilities}
   */
  static getUtilities() {
    return Utilities;
  }

  /**
   * Retrieve the configuration instance.
   *
   * @param {string} namespace
   * @returns {Config}
   * @throws {Error}
   */
  getConfig(namespace) {
    if (namespace) {
      if (typeof namespace === 'string') {
        return this._pluginManager.getPluginByName(namespace).getConfig();
      } else {
        throw new Error('{namespace} parameter must be of type string and reference a plugin')
      }
    }
    return this._config;
  }

  /**
   * Check if the specified configuration file is already
   * published to the config folder.
   *
   * @param {Config} config
   * @param {string} fileName
   * @returns {boolean}
   */
  isConfigurationFilePublished(config, fileName) {
    if (config instanceof Config) {
      return Utilities.isFilePublished(fileName, config.getConfigDirectory());
    } else {
      throw new Error('{config} must be an instance of {Config}');
    }
  }

  /**
   * Publish the configuration file with the specified name
   * into the specified config folder.
   *
   * @param {Config} config
   * @param {string} filePath
   */
  publishConfigurationFile(config, filePath) {
    if (config instanceof Config) {
      return Utilities.publishFile(filePath, config.getConfigDirectory());
    } else {
      throw new Error('{config} must be an instance of {Config}');
    }
  }

  /**
   * Retrieve the root directory.
   *
   * @returns {string}
   */
  getRootDirectory() {
    return this._rootDir;
  }

  /**
   * Retrieve the path to the views directory.
   *
   * @returns {string}
   */
  getViewsDirectory() {
    return this._viewsDir;
  }

  /**
   * Retrieve the path to the public directory.
   *
   * @returns {string}
   */
  getPublicDirectory() {
    return this._publicDir;
  }

  /**
   * Retrieve the path to the storage directory.
   *
   * @returns {string}
   */
  getStorageDirectory() {
    return this._storageDir;
  }

  /**
   * Retrieve the socket.io instance.
   *
   * @returns {Socket}
   */
  getWebSocketHandle() {
    return this._io;
  }

  /**
   * Retrieve the main {Express.Router} instance.
   *
   * @returns {Express.Router}
   */
  getRouter() {
    return this._router;
  }

  /**
   * Create a router with a context applied to each request.
   *
   * @param {object} options
   * @return {Express.Router}
   */
  static createRouter(options) {
    let router = Express.Router();
    Quture.contextualize(router);

    router.use(function(request, response, next) {
      router.generateContext(request);
      next();
    });

    return router;
  }

  /**
   * Retrieve the base error types.
   *
   * @returns {object}
   */
  static get ErrorTypes() {
    return ErrorTypes;
  }

  /**
   * Retrieve the {ErrorResponseFormatter} instance.
   *
   * @returns {ErrorResponseFormatter}
   */
  getErrorResponseFormatter() {
    return this._errorResponseFormatter;
  }

  /**
   * Create a new {ErrorFormatter} instance.
   *
   * @returns {ErrorFormatter}
   */
  static createErrorFormatter(errorType, errorFormatter) {
    return new ErrorFormatter(errorType, errorFormatter);
  }

  /**
   * Create a new {ErrorCode} instance.
   *
   * @returns {ErrorCode}
   */
  static createErrorCode(code, meta) {
    return new ErrorCode(code, meta);
  }

  /**
   * Apply a context to the specified object.
   *
   * @param {object} object
   */
  static contextualize(object, namespace) {
    if (namespace) {
      if (typeof namespace === 'string') {
        object._ctxId = namespace;
      } else {
        throw new Error('{namespace} must be of type {string}');
      }
    } else {
      object._ctxId = RandomString.generate(5);
    }

    /**
     * Retrieve context ID.
     *
     * @returns {string}
     */
    object.getContextId = function() {
      return object._ctxId;
    }

    /**
     * Instantiate a new Context and apply it to the specified object.
     *
     * @param {object} k
     */
    object.generateContext = function(k) {
      if (object.getContext(k)) {
        throw new Error('A context already exists for the specified object');
      } else {
        let context = new Context();
        k['q-ctx-' + object.getContextId()] = context;
        return context;
      }
    }

    /**
     * Retrieve context from the specified object.
     *
     * @returns {Context}
     */
    object.getContext = function(k) {
      return k['q-ctx-' + object.getContextId()];
    }
  }

  /**
   * Bootstrap the application.
   *
   * @returns {Express}
   */
  bootstrap() {
    const self = this;

    if (!this._isBootstrapped) {
      self._isBootstrapped = true;

      self._app = Express();
      self._app.disable('etag');

      /**
       * Configure the application proxy settings
       */
      self._app.set('trust proxy', self.getConfig().get('app.behindProxy', false));

      /**
       * Configure the application port
       */
      self._app.set('port', function(val) {
        var port = parseInt(val, 10);
        if (isNaN(port)) {
          return val;
        }
        if (port >= 0) {
          return port;
        }
        return false;
      }(self.getConfig().get('app.port', 80)));

      /**
       * Setup view engine && pretty responses
       */
      if (self.getConfig().get('app.prettify', false)) {
        self._app.locals.pretty = true;
        self._app.set('json spaces', 2);
      }

      self._app.set('views', self.getViewsDirectory());
      self._app.set('view engine', self.getConfig().get('app.viewEngine', 'pug'));

      /**
       * Serve the favicon
       */
      let faviconPath = path.join(self.getPublicDirectory(), 'favicon.ico');
      if (fs.existsSync(faviconPath)) {
        self._app.use(favicon(faviconPath));
      }

      /**
       * ... Middleware ...
       */
      if (process.env.NODE_ENV !== 'testing') {
        self._app.use(logger('dev'));
      }

      self._app.use(bodyParser.json());
      self._app.use(bodyParser.urlencoded({ 'extended': true }));
      self._app.use(cookieParser(self.getConfig().get('app.key', null), {
        'httpOnly': true,
        'secure': self.getConfig().get('app.secure', false)
      }));
      self._app.use(helmet());
      self._app.use(methodOverride('X-HTTP-Method-Override'));
      self._app.use(cors());

      // Append a method to retrieve the global context
      self._app.use(function(request, response, next) {
        request.quture = self;
        self.generateContext(request);

        /**
         * Retrieve the global context.
         *
         * @return {Context}
         */
        request.getGlobalContext = function() {
          return self.getContext(request);
        }

        next();
      });

      /**
       * Configure non-api endpoints
       */
      self._app.use('/public', Express.static(this.getPublicDirectory()));

      // Apply router after the middleware has been applied
      self._app.use(self._router);

      /**
       * Redirect other calls to the application router, if a route cannot be found
       * or the error isn't handled, then fallback to the public directory thus
       * leaving the client to deal with it.
       */
      self._app.use(function(request, response, next) {
        if (request.originalUrl.startsWith('/api')) {
          next(new ErrorTypes.NotFoundError());
        } else {
          let indexFile = path.join(self.getPublicDirectory(), 'index.html');
          if (fs.existsSync(indexFile)) {
            response.sendFile(indexFile);
          } else {
            next();
          }
        }
      }, Express.static(self.getPublicDirectory()));

      // Apply 404 catch
      self._app.use(function(request, response, next) {
        next(new ErrorTypes.NotFoundError()); // Resource not found
      });

      // Apply a route response handler
      self._app.use(function(error, request, response, next) {
        debug(error);
        response.status(error.status || 200).json({
          errors: [
            self.getErrorResponseFormatter().format(error)
          ]
        });
      });
    }

    return self._app;
  }

}
