'use strict';

import chai from 'chai';
import path from 'path';
import fsx from 'fs-extra';
import IO from 'socket.io';
import Express from 'express';
import chaiHttp from 'chai-http';
import Quture from '../src/quture';
import Config from 'quture-config';
import Plugin from 'quture-plugin';
import ErrorTypes from '../src/errors';
import ErrorResponseFormatter from '../src/errors/error-response-formatter';

describe('quture-app', function() {

  let rootDir = path.join(__dirname);
  let configPath = path.join(__dirname, 'config');
  let viewsPath = path.join(__dirname, 'views');
  let publicPath = path.join(__dirname, 'public');
  let storagePath = path.join(__dirname, 'storage');

  /**
   * Setup chai
   */
  before(function(done) {
    chai.use(chaiHttp);
    fsx.removeSync(configPath);
    fsx.removeSync(viewsPath);
    fsx.removeSync(publicPath);
    fsx.removeSync(storagePath);

    fsx.ensureDirSync(configPath);
    done();
  });

  /**
   * Tests
   */

  it ('shouldn\'t create a new {Quture} instance without a valid root directory', function(done) {
    try {
      let instance = new Quture();
      done(new Error('A new {Quture} instance was created'));
    } catch (error) {
      if (error.message.startsWith('{options.rootDir}')) {
        done();
      } else {
        throw error;
      }
    }
  });

  it ('should create a new {Quture} instance', function(done) {
    try {
      let instance = new Quture({ rootDir });
      done();
    } catch (error) {
      done(error);
    }
  });

  it('should retrieve the {Config} instance', function(done) {
    let instance = new Quture({ rootDir });

    let config = instance.getConfig();

    if (config && config instanceof Config) {
      done();
    } else {
      done(new Error('Unable to retrieve the {Config} instance'));
    }
  });

  it('should retrieve the specified public directory', function(done) {
    let instance = new Quture({ rootDir });

    let dir = instance.getPublicDirectory();

    if (dir === publicPath) {
      done();
    } else {
      done(new Error('The public directory is different to the one returned'));
    }
  });

  it('should retrieve the {Socket.IO} instance', function(done) {
    let instance = new Quture({ rootDir });

    let io = instance.getWebSocketHandle();

    if (io && io instanceof IO) {
      done();
    } else {
      done(new Error('Unable to retrieve the {Socket.IO} instance'));
    }
  });

  it('should retrieve the base error types', function(done) {
    let instance = new Quture({ rootDir });

    let errorTypes = Quture.ErrorTypes;

    if (errorTypes && errorTypes === ErrorTypes) {
      done();
    } else {
      done(new Error('Unable to retrieve the base error types'));
    }
  });

  it('should retrieve the {ErrorResponseFormatter} instance', function(done) {
    let instance = new Quture({ rootDir });

    let errorResponseFormatter = instance.getErrorResponseFormatter();

    if (errorResponseFormatter && errorResponseFormatter instanceof ErrorResponseFormatter) {
      done();
    } else {
      done(new Error('Unable to retrieve the {ErrorResponseFormatter} instance'));
    }
  });

  it('should retrieve the {Router} instance', function(done) {
    let instance = new Quture({ rootDir });

    let router = instance.getRouter();

    if (router) {
      done();
    } else {
      done(new Error('Unable to retrieve the {Router} instance'));
    }
  });

  it('should be able to retrieve the {Express} instance if the application has been bootstrapped', function(done) {
    let instance = new Quture({ rootDir });

    try {
      let app = instance.bootstrap();

      done();
    } catch (error) {
      done(error);
    }

  });

  it('should throw a 404 error for a basic bootstrapped application', function(done) {
    let instance = new Quture({ rootDir });

    chai.request(instance.bootstrap())
      .get('/')
      .end(function (error, response) {
        chai.expect(response).to.be.json;
        chai.expect(response.body).to.have.property('errors');
        chai.expect(response.body.errors).to.be.an('array');
        chai.expect(response.body.errors).to.have.length.above(0);
        chai.expect(response.body.errors[0]).to.have.property('type');
        chai.expect(response.body.errors[0].type).to.equal(ErrorTypes.NotFoundError.getType());
        chai.expect(response).to.have.status(404);
        done();
      });
  });

  it('should throw a 404 error for an unknown route', function(done) {
    let instance = new Quture({ rootDir });

    chai.request(instance.bootstrap())
      .get('/q/unknown/resource')
      .end(function (error, response) {
        chai.expect(response).to.be.json;
        chai.expect(response.body).to.have.property('errors');
        chai.expect(response.body.errors).to.be.an('array');
        chai.expect(response.body.errors).to.have.length.above(0);
        chai.expect(response.body.errors[0]).to.have.property('type');
        chai.expect(response.body.errors[0].type).to.equal(ErrorTypes.NotFoundError.getType());
        chai.expect(response).to.have.status(404);
        done();
      });
  });

  // TODO: remove this once graphql is implemented
  it('should throw a 404 error for an api route', function(done) {
    let instance = new Quture({ rootDir });

    chai.request(instance.bootstrap())
      .get('/api/unknown/resource')
      .end(function (error, response) {
        chai.expect(response).to.be.json;
        chai.expect(response.body).to.have.property('errors');
        chai.expect(response.body.errors).to.be.an('array');
        chai.expect(response.body.errors).to.have.length.above(0);
        chai.expect(response.body.errors[0]).to.have.property('type');
        chai.expect(response.body.errors[0].type).to.equal(ErrorTypes.NotFoundError.getType());
        chai.expect(response).to.have.status(404);
        done();
      });
  });

  // TODO: more GraphQL tests & Plugin tests

});
